/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 * contact: sp@utia.cz
 *
 * Filename: main.c
 * Date: 29.06.2016
 * Author: Lukas Kohout <kohoutl@utia.cas.cz>
 * Description:
 *
 * Changelog:
 *
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <signal.h>


#include <xparameters.h>

// ***************************************************************************
#include "ui.h"

// application menu
const ui_menu_t menu[] = {
  {'1', "Set HDMI out 1920x1080p60"     },
  {'2', "Set HDMI out 1920x1200p50"     },
  {'3', "Set HDMI out 1600x1200p50"     },
  {'4', "Set HDMI out 1680x1050p60"     },
  {'5', "Set HDMI out 1280x1024p60"     },
  {'6', "Set HDMI out 1280x720p60"      },
  {'7', "Set HDMI out 1024x768p60"      },
  {'8', "Set HDMI out 800x600p60"       },
  {'9', "Set HDMI out 640x480p60"       },
  {'a', "Set HDMI out 600x800p60"       },

  {'l', "Shift HDMI out left"           },
  {'r', "Shift HDMI out right"          },
  {'d', "Shift HDMI out down"           },
  {'u', "Shift HDMI out up"             },
  {'z', "Set HDMI out offset to [0, 0]" },

  {'m', "Print this menu"               },
  {'x', "Quit"                          },
  {  0, NULL                            }
};

#include <termios.h>
struct termios initial_settings, new_settings;


// ***************************************************************************
#define I2C_BASE  XPS_I2C0_BASEADDR
#define DDR_MEM_BASE  XPAR_DDR_MEM_BASEADDR


// ***************************************************************************
// Video Timing Controller VTC
#include "vtc.h"

vtc_t vtc;
extern timing_t vtc_resolutions[VTC_RES_CNT];

// ***************************************************************************
// Video DMA VDMA
#include "video_frame_buffer.h"

#define MAX_RES_X       1920
#define MAX_RES_Y       1200 //1080
#define MAX_FRAME_SIZE  2048*4096              // max size of one frame for VDMA aligned to kB

#define FB_OFFSET       0x2D000000  // physical address of framebuffer reserved memory
#define FB_BASE         DDR_MEM_BASE + FB_OFFSET

video_resolution_t  fb_res_in;
video_resolution_t  fb_res_out;
video_resolution_t  fb_store_res;

/*
 * Frame offset
 */
video_offset_t offset = {0, 0};

#define OFFSET_EN_NONE  0
#define OFFSET_EN_ZERO  1
#define OFFSET_EN_LEFT  2
#define OFFSET_EN_RIGHT 3
#define OFFSET_EN_DOWN  4
#define OFFSET_EN_UP    5
unsigned int offset_en = OFFSET_EN_NONE;


// ***************************************************************************
// IMAGEON
#include "fmc_iic.h"
#include "fmc_imageon.h"
#include "fmc_imageon_iic.h"
#include "fmc_imageon_hdmi_in.h"
#include "fmc_imageon_hdmi_framebuffer.h"

fmc_imageon_hdmi_framebuffer_t fb;

/*
 * EDID content to provide PC input properties to the PC input source
 */
extern u8 utia_hdmii_edid_content[256];


// ***************************************************************************
// Simple hot plug of HDMII
typedef struct {
    unsigned int is_cable_detected;
    unsigned int is_video_valid;
    unsigned int is_unplugged;
} hot_plug_status_t;

hot_plug_status_t hdmii_HPS;


// ***************************************************************************
// programmble clock wizard
#include "clkwiz.h"
#define CLKWIZ_BASE   XPAR_CLK_GEN_CLKGEN_VARIABLE_BASEADDR

clkwiz_t clkwiz;


// ***************************************************************************

int system_stop = 0;

// signal handler for CTRL+c
void sig_handler(int signo)
{
    if (signo == SIGINT) {
        printf("Received SIGINT\n");
        system_stop = 1;
    }
}

// ***************************************************************************
// ***************************************************************************

// #define FATAL do { fprintf(stderr, "Error at line %d, file %s (%d) [%s]\n", \
// __LINE__, __FILE__, errno, strerror(errno)); exit(-1); } while(0)

// ***************************************************************************
// memory mapping support (peripherals)

#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)

typedef struct {
    void *base;
    void *dev_base;
} mem_map_t;

int mapio(int memfd, off_t dev_base, mem_map_t *m)
{
    // Map one page of memory into user space such that the device is in that page, but it may not
    // be at the start of the page.

    m->base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, dev_base & ~MAP_MASK);
    if ( m->base == MAP_FAILED ) {  // MAP_FAILED = (void *) -1
        return XST_FAILURE;
    }
    // printf("Memory mapped at address %p.\n", m->base);

    // get the address of the device in user space which will be an offset from the base
    // that was mapped as memory is mapped at the start of a page

    m->dev_base = (void*)((char*)(m->base) + (dev_base & MAP_MASK));

    return XST_SUCCESS;
}


/*
 * Framebuffer memory mapping (assumes each frame start aligned to 4096 bytes)
 */
unsigned mapfb(int memfd, off_t base, mem_map_t *m, unsigned long map_size)
{
    // Map one page of memory into user space such that the device is in that page, but it may not
    // be at the start of the page.

    if ((base % MAP_SIZE) != 0) {
        printf("Start of memory block must be aligned to 4096 bytes\r\n");
        return XST_FAILURE;
    }

    m->base = mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, base);
    if ( m->base == MAP_FAILED ) {
        return XST_FAILURE;
    }
    //printf("Memory mapped at address %p.\n", m->base);

    // get the address of the device in user space which will be an offset from the base
    // that was mapped as memory is mapped at the start of a page

    m->dev_base = m->base;

    return XST_SUCCESS;
}


// ***************************************************************************
// define black color in YCrCb 4:2:2 color space
#define PIX_Y_INIT     0x00
#define PIX_CBCR_INIT  0x80

#define PIX_Y_WHITE     0xFF
#define PIX_CBCR_WHITE  0x80

int clear_fb_YCrCb(fmc_imageon_hdmi_framebuffer_t *fb, int memfd)
{
    mem_map_t fb_map;
    volatile u8 *pStorageMem;
    int i;

    if ( mapfb(memfd, FB_BASE, &fb_map, fb->uNumFrames_HdmiFrameBuffer * MAX_FRAME_SIZE) ) {
        printf("Frame buffer mapping to userspace failed\n");
        return XST_FAILURE;
    }

    pStorageMem = (u8 *)fb_map.dev_base;
    for ( i = 0; i < fb->uNumFrames_HdmiFrameBuffer * MAX_FRAME_SIZE; i += fb->vdma_rx_bytes ) {
        // Init Pixels
        *pStorageMem++ = PIX_Y_INIT;
        *pStorageMem++ = PIX_CBCR_INIT;
    }

    if ( munmap(fb_map.base, MAP_SIZE) == -1 ) {
        printf("Frame buffer userspace unmapping failed\n");
        return XST_FAILURE;
    }

    return XST_SUCCESS;
}


// ***************************************************************************
int set_resolution(clkwiz_t *clkwiz, vtc_t *vtc, fmc_imageon_hdmi_framebuffer_t *fb, u8 sel, video_resolution_t *res)
{
    int status = 0;
    u32 timeout;

    /*
     * set new pixel clock
     */
    if (sel == VTC_600x800p60)
        clkwiz_clkout0_lookup(&(clkwiz->clkout0), VTC_800x600p60);
    else
        clkwiz_clkout0_lookup(&(clkwiz->clkout0), sel);

    printf("Initialize Clock Wizard ... ");

    clkwiz_set_vco(clkwiz);
    clkwiz_set_clkout0(clkwiz);
    clkwiz_update(clkwiz->base_addr);

    timeout = 100;
    status = 0;
    while (  !status && timeout-- ) {
        status = is_clkwiz_locked(clkwiz->base_addr);
        if (!status) usleep(10);
    }
    if (!status) {
        printf("FAILED\n");
        return XST_FAILURE;
    }
    printf("OK\n");

    /*
     * Initialize VDMA TX
     */
    printf("Initialize VDMA ... ");
    status = vfb_tx_stop(&fb->vdma_hdmi);
    if (status != XST_SUCCESS) {
        printf("FAILED - Cannot stop VDMA 1 TX\n");
        return XST_FAILURE;
    }

    res->x = vtc_resolutions[sel].HActiveVideo;
    res->y = vtc_resolutions[sel].VActiveVideo;

    printf("TX Init ... ");
    status = vfb_tx_init( &fb->vdma_hdmi,                     // pAxiVdma
                          &fb->vdmacfg_hdmi_read,             // pReadCfg
                          res,                                // uVideoResolution
                          &fb_store_res,                      // uStorageResolution  -- always set to maximum
                          fb->uBaseAddr_MEM_HdmiFrameBuffer,  // uMemAddr
                          fb->uNumFrames_HdmiFrameBuffer,     // uNumFrames
                          &offset,                            // horizontal and vertical offset
                          fb->vdma_tx_bytes                   // number of bytes per pixel
    );
    if (status != XST_SUCCESS) {
        printf("FAILED\n");
        return XST_FAILURE;
    } else {
        printf("OK\n");
    }

    /*
     * initialize IMAGEON HDMIO
     */
    printf("HDMIO Initialization ... " );
    fb->hdmio_timing.IsHDMI = 0;

    fb->hdmio_timing.HActiveVideo  = vtc_resolutions[sel].HActiveVideo;
    fb->hdmio_timing.HFrontPorch   = vtc_resolutions[sel].HFrontPorch;
    fb->hdmio_timing.HSyncWidth    = vtc_resolutions[sel].HSyncWidth;
    fb->hdmio_timing.HBackPorch    = vtc_resolutions[sel].HBackPorch;
    fb->hdmio_timing.HSyncPolarity = vtc_resolutions[sel].HSyncPolarity;

    fb->hdmio_timing.VActiveVideo  = vtc_resolutions[sel].VActiveVideo;
    fb->hdmio_timing.VFrontPorch   = vtc_resolutions[sel].VFrontPorch;
    fb->hdmio_timing.VSyncWidth    = vtc_resolutions[sel].VSyncWidth;
    fb->hdmio_timing.VBackPorch    = vtc_resolutions[sel].VBackPorch;
    fb->hdmio_timing.VSyncPolarity = vtc_resolutions[sel].VSyncPolarity;

    status = fmc_imageon_hdmio_init(&(fb->fmc_imageon),
                                    1,                    // hdmioEnable = 1
                                    &(fb->hdmio_timing),  // pTiming
                                    1                     // waitHPD = 0
    );
    if (!status) // fmc_imageon_hdmio_init return 1 on success
        printf("FAILED\n");
    else
        printf("OK\n");


    /*
     * initialize timing controller
     */
    printf("Initialize Timing Controller ... ");
    vtc->timing_sel = sel;

    XVtc_Disable(&vtc->Vtc);

    fill_vtc_timing(vtc);
    set_vtc_timing(vtc);

    printf(" %s ", vtc_resolutions[sel].pName);

    printf(" OK\n");

    return XST_SUCCESS;
}


/*
 * Checks the event cable plugged (PC input) and initialize HDMI input
 */
void check_hdmii_cable_plugged(hot_plug_status_t *hp, fmc_imageon_t *inst, u8 edid[256])
{
    int status = 0;

    const u8 llc_delay_space  = (0x98 >> 1);
    const u8 llc_delay_reg    = 0x19;
    u8 llc_delay_val    = 0x80;  // no LLC delay

    if ( !hp->is_cable_detected && fmc_imageon_hdmii_cable_detect(inst) ) {
        hp->is_cable_detected = 1;
        hp->is_video_valid = 0;

        // Initialize HDMI Input, release reset & set HPD (including EDID content)
        // returns 1 on success
        status = fmc_imageon_hdmii_init(inst,
                                        1,      // hdmiiEnable = 0
                                        1,      // editInit = 1
                                        edid);

        // set LLC DLL phase = 0, let INV_LLC_POL = 1
        // returns number of bytes
        status += fb.fmc_imageon_iic.fpIicWrite(&(fb.fmc_imageon_iic), llc_delay_space, llc_delay_reg, &llc_delay_val, 1);

        if ( status != 2 ) {
            printf( "ERROR : Failed to initialize HDMI Input Interface\n" );
            return;
        }

        printf("HDMII cable plugged\n");
    }
}

/*
 * Checks the event cable unplugged (PC input) and initialize HDMI input
 */
void check_hdmii_cable_unplugged(hot_plug_status_t *hp, fmc_imageon_t *inst)
{
    if ( hp->is_cable_detected && !fmc_imageon_hdmii_cable_detect(inst) ) {
        // cable unplugged (not powered)
        hp->is_cable_detected = 0;
        hp->is_video_valid = 0;
        hp->is_unplugged = 1;
        printf("HDMII cable unplugged\n");
    }
}

/*
 * Detects change of incomig input signals and trigger the reinitialization of PC input
 */
void check_hdmii_resolution(hot_plug_status_t *hp, fmc_imageon_hdmi_framebuffer_t *fb, video_resolution_t *res)
{
    if (hp->is_video_valid) {
        (void)fmc_imageon_hdmii_get_timing( &(fb->fmc_imageon), &(fb->hdmii_timing) );
        fb->hdmii_width  = fb->hdmii_timing.HActiveVideo;
        fb->hdmii_height = fb->hdmii_timing.VActiveVideo;
        if (
            (res->x   != fb->hdmii_width)                 || (res->y   != fb->hdmii_height)               ||
            (res->hfp != fb->hdmii_timing.HFrontPorch)    || (res->vfp != fb->hdmii_timing.VFrontPorch)   ||
            (res->hsw != fb->hdmii_timing.HSyncWidth)     || (res->vsw != fb->hdmii_timing.VSyncWidth)    ||
            (res->hbp != fb->hdmii_timing.HBackPorch)     || (res->vbp != fb->hdmii_timing.VBackPorch)    ||
            (res->hsp != fb->hdmii_timing.HSyncPolarity)  || (res->vsp != fb->hdmii_timing.VSyncPolarity)
        ) {
            hp->is_video_valid = 0;
        }
    }
}

/*
 * Set VDMA input to valid video data from HDMII
 */
void set_valid_hdmii_video_signal(hot_plug_status_t *hp, fmc_imageon_hdmi_framebuffer_t *fb, video_resolution_t *res, int memfd)
{
    int status;

    if (hp->is_cable_detected && !hp->is_video_valid) {

        get_hdmii_timing(fb);
        print_hdmii_timing(&(fb->hdmii_timing));

        if (fb->hdmii_timing.IsInterlaced) {
          printf("\n\tERROR : Interlaced video signal not supported!\n");
          return;
        }

        if (fb->hdmii_timing.IsEncrypted) {
          printf("\n\tERROR : HDCP encrypted video signal not supported!\n");
          return;
        }

        printf("Initialize VDMA ... ");
        vfb_rx_stop( &(fb->vdma_hdmi) );

        // Clear frame stores
        if ( clear_fb_YCrCb(fb, memfd) ) return;

        // set currently locked resolution
        res->x    = fb->hdmii_timing.HActiveVideo;
        res->y    = fb->hdmii_timing.VActiveVideo;

        res->hfp  = fb->hdmii_timing.HFrontPorch;
        res->hsw  = fb->hdmii_timing.HSyncWidth;
        res->hbp  = fb->hdmii_timing.HBackPorch;
        res->hsp  = fb->hdmii_timing.HSyncPolarity;

        res->vfp  = fb->hdmii_timing.VFrontPorch;
        res->vsw  = fb->hdmii_timing.VSyncWidth;
        res->vbp  = fb->hdmii_timing.VBackPorch;
        res->vsp  = fb->hdmii_timing.VSyncPolarity;

        // initialize VDMA input
        printf("RX Init ... ");
        status = vfb_rx_init( &(fb->vdma_hdmi),                   // pAxiVdma
                              &(fb->vdmacfg_hdmi_write),          // pWriteCfg
                              res,                                // uVideoResolution     -- use currently locked resolution
                              &fb_store_res,                      // uStorageResolution   -- always set to maximum
                              fb->uBaseAddr_MEM_HdmiFrameBuffer,  // uMemAddr
                              fb->uNumFrames_HdmiFrameBuffer,     // uNumFrames
                              fb->vdma_rx_bytes                   // number of bytes per pixel
        );

        if (status != XST_SUCCESS) {
            printf( "FAILED\n\n" );
            return;
        } else {
            printf( "OK\n\n" );
        }

        // resolution changes have impact on errors flags - clear them
        vfb_clr_errors(&(fb->vdma_hdmi));

        hp->is_video_valid = 1;

    }
}


/*
 * sets required video offset
 */
void set_hdmio_offset(hot_plug_status_t *hp, fmc_imageon_hdmi_framebuffer_t *fb,
                      video_resolution_t *r1, video_resolution_t *r2,
                      unsigned int *off_en, video_offset_t *off)
{
    if (hp->is_video_valid && *off_en) {
        switch ( *off_en ) {
            case OFFSET_EN_ZERO :
                off->x = 0;
                off->y = 0;
                break;
            case OFFSET_EN_LEFT :
                if ((int)off->x < (int)(r2->x - r1->x)) off->x += 4;  // 4 pixels * 2 bytes -> 64-bit tranfer
                break;
            case OFFSET_EN_RIGHT :
                if (off->x > 0) off->x -= 4;  // 4 pixels * 2 bytes -> 64-bit tranfer
                break;
            case OFFSET_EN_DOWN :
                if ((int)off->y < (int)(r2->y - r1->y)) off->y++;
                break;
            case OFFSET_EN_UP :
                if (off->y > 0) off->y--;
                break;
            default : break;
        }

        *off_en = OFFSET_EN_NONE;

        (void)vfb_tx_init(  &fb->vdma_hdmi,                     // pAxiVdma
                            &fb->vdmacfg_hdmi_read,             // pReadCfg
                            r1,                                 // uVideoResolution
                            &fb_store_res,                      // uStorageResolution  -- always set to maximum
                            fb->uBaseAddr_MEM_HdmiFrameBuffer,  // uMemAddr
                            fb->uNumFrames_HdmiFrameBuffer,     // uNumFrames
                            off,                                // horizontal and vertical offset
                            fb->vdma_tx_bytes                   // number of bytes per pixel
        );


        printf("Frame offset: [%d, %d]\n", (unsigned int)off->x, (unsigned int)off->y);
    }
}


// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
int main()
{
    int status;

    int keyin;
    u8 key;
    u8 menu_print_en = 1;
    u32 timeout;


    printf("\n\n*****************************************************************************\n");
    printf(    "*  Signal Processing Dept., UTIA AV CR, v.v.i.                              *\n");
    printf(    "*  HDMI In/Out IMAGEON Application                                          *\n");
    printf(    "*****************************************************************************\n\n");


    /*
     * open memory device
     */
    int memfd;
    memfd = open("/dev/mem", O_RDWR | O_SYNC);
    if (memfd == -1) {
        printf("Can't open /dev/mem.\n");
        return XST_FAILURE;
    }
    //printf("/dev/mem opened.\n");


    /*
     * set clock wizard
     */
    printf("Initialize Clock Wizard ... ");

    mem_map_t clkwiz_map;
    if ( mapio(memfd, CLKWIZ_BASE, &clkwiz_map) ) {
        printf("Clock Wizard mapping to userspace failed\n");
        return XST_FAILURE;
    }

    clkwiz.base_addr    = (u32)clkwiz_map.dev_base;
    clkwiz.clk_in       = 50.0;         //50 MHz
    clkwiz_clkout0_lookup(&(clkwiz.clkout0), CLKWIZ_148_5_MHZ);

    clkwiz_set_vco(&clkwiz);
    clkwiz_set_clkout0(&clkwiz);
    clkwiz_update(clkwiz.base_addr);

    timeout = 100;
    status = 0;
    while (  !status && timeout-- ) {
        status = is_clkwiz_locked(clkwiz.base_addr);
        if (!status) usleep(10);
    }
    if ( !status ) {
        printf("FAILED\n");
        return XST_FAILURE;
    }
    printf("OK\n");


    /*
     * HDMI output - IMAGEON on FMC
     */
    mem_map_t iic_map;
    if ( mapio(memfd, XPAR_FMC2_IMAGEON_IIC_BASEADDR, &iic_map) ) {
        printf("Master IIC mapping to userspace failed\n");
        return XST_FAILURE;
    }
    fb.uBaseAddr_IIC_FmcImageon       = (u32) iic_map.dev_base;
    fb.uDeviceId_VTC_HdmiiDetector    = 0; // no detector used in vdma branch
    fb.uDeviceId_VTC_HdmioGenerator   = 0; // no generator needed in vdma branch
    fb.bVerbose                       = 0;


    /*
     * Initialize I2C for FMC
     */
    printf("IIC IMAGEON Initialization ... " );
    status = fmc_iic_xps_init(&(fb.fmc_imageon_iic), "\tHDMII-IMAGEON IIC Controller", fb.uBaseAddr_IIC_FmcImageon );
    if ( !status ) {
        printf( "FAILED\n" );
        return XST_FAILURE;
    }

    status = fmc_imageon_init(&(fb.fmc_imageon), "\tHDMII-IMAGEON PC Input", &(fb.fmc_imageon_iic));
    if (!status) {  // fmc_imageon_hdmio_init return 1 on success
        printf("FAILED\n");
        return XST_FAILURE;
    } else {
        printf("OK\n");
    }

    fb.fmc_imageon.bVerbose = fb.bVerbose;


    /*
     * Initialize IMAGEON HDMI Output
     */
    printf("HDMIO Initialization ... " );
    fb.hdmio_timing.IsHDMI = 0;

    fb.hdmio_timing.HActiveVideo  = vtc_resolutions[VTC_1920x1080p60].HActiveVideo;
    fb.hdmio_timing.HFrontPorch   = vtc_resolutions[VTC_1920x1080p60].HFrontPorch;
    fb.hdmio_timing.HSyncWidth    = vtc_resolutions[VTC_1920x1080p60].HSyncWidth;
    fb.hdmio_timing.HBackPorch    = vtc_resolutions[VTC_1920x1080p60].HBackPorch;
    fb.hdmio_timing.HSyncPolarity = vtc_resolutions[VTC_1920x1080p60].HSyncPolarity;

    fb.hdmio_timing.VActiveVideo  = vtc_resolutions[VTC_1920x1080p60].VActiveVideo;
    fb.hdmio_timing.VFrontPorch   = vtc_resolutions[VTC_1920x1080p60].VFrontPorch;
    fb.hdmio_timing.VSyncWidth    = vtc_resolutions[VTC_1920x1080p60].VSyncWidth;
    fb.hdmio_timing.VBackPorch    = vtc_resolutions[VTC_1920x1080p60].VBackPorch;
    fb.hdmio_timing.VSyncPolarity = vtc_resolutions[VTC_1920x1080p60].VSyncPolarity;

    status = fmc_imageon_hdmio_init(&(fb.fmc_imageon),
                                    1,                  // hdmioEnable = 1
                                    &(fb.hdmio_timing), // pTiming
                                    1                   // waitHPD = 0
    );
    if ( !status ) {  // fmc_imageon_hdmio_init return 1 on success
        printf("FAILED\n");
        return XST_FAILURE;
    } else {
        printf("OK\n");
    }


    /*
     * initialize timing controller
     */
    printf("Initialize Timing Controller ... ");
    vtc.timing_sel = VTC_1920x1080p60;

    vtc.VtcCfgPtr = XVtc_LookupConfig(XPAR_VTC_0_DEVICE_ID);
    if (vtc.VtcCfgPtr == NULL) {
        printf("Failed to get VTC config\n");
        return XST_FAILURE;
    }

    mem_map_t vtc_map;
    if ( mapio(memfd, vtc.VtcCfgPtr->BaseAddress, &vtc_map) ) {
        printf("VTC mapping to userspace failed\n");
        return XST_FAILURE;
    }
    vtc.VtcCfgPtr->BaseAddress = (u32) vtc_map.dev_base;

    status = XVtc_CfgInitialize(&vtc.Vtc, vtc.VtcCfgPtr, vtc.VtcCfgPtr->BaseAddress);
    if (status != XST_SUCCESS) {
        printf("Failed to init VTC");
        return XST_FAILURE;
    }

    XVtc_Disable(&vtc.Vtc);

    fill_vtc_timing(&vtc);
    set_vtc_timing(&vtc);

    printf(" %s ", vtc_resolutions[VTC_1920x1080p60].pName);

    printf(" OK\n");


    /*
     * initialize VDMA output to display FB
     */
    printf("Initialize VDMA ... ");

    fb.uDeviceId_VDMA_HdmiFrameBuffer  = XPAR_AXIVDMA_0_DEVICE_ID;
    fb.uBaseAddr_MEM_HdmiFrameBuffer   = FB_BASE;                    // frame storage for vdma
    fb.uNumFrames_HdmiFrameBuffer      = XPAR_AXIVDMA_0_NUM_FSTORES;
    fb.vdma_tx_bytes                   = 2;
    fb.vdma_rx_bytes                   = 2;


    fb_store_res.x  = MAX_RES_X;  // always set to maximum
    fb_store_res.y  = MAX_RES_Y;  // always set to maximum
    fb_res_out.x    = vtc_resolutions[VTC_1920x1080p60].HActiveVideo;
    fb_res_out.y    = vtc_resolutions[VTC_1920x1080p60].VActiveVideo;

    printf("Common Init ... ");
    XAxiVdma_Config *Config;
    mem_map_t vdma_map;

    Config = XAxiVdma_LookupConfig( fb.uDeviceId_VDMA_HdmiFrameBuffer );
    if (!Config) {
        printf( "No video DMA found for ID %d\n", fb.uDeviceId_VDMA_HdmiFrameBuffer);
        return 1;
    }

    if ( mapio(memfd, Config->BaseAddress, &vdma_map) ) {
        printf("VDMA mapping to userspace failed\n");
        return XST_FAILURE;
    }

    /*
     * update baseAddress pointer after mapping
     */
    Config->BaseAddress = (u32) vdma_map.dev_base;


    /* Initialize DMA engine */
    status = XAxiVdma_CfgInitialize(&fb.vdma_hdmi, Config, Config->BaseAddress);
    if (status != XST_SUCCESS) {
      printf( "Initialization failed %d\n", status);
      return 1;
    }

    printf("TX Init ... ");
    status = vfb_tx_init( &fb.vdma_hdmi,                    // pAxiVdma
                          &fb.vdmacfg_hdmi_read,            // pReadCfg
                          &fb_res_out,                      // uVideoResolution
                          &fb_store_res,                    // uStorageResolution  -- always set to maximum
                          fb.uBaseAddr_MEM_HdmiFrameBuffer, // uMemAddr
                          fb.uNumFrames_HdmiFrameBuffer,    // uNumFrames
                          &offset,                          // horizontal and vertical offset
                          fb.vdma_tx_bytes                  // number of output bytes
    );
    if (status != XST_SUCCESS) {
        printf("FAILED\n");
        return XST_FAILURE;
    } else {
        printf("OK\n");
    }


    /*
     * clear HDMII flags and params before autodetection
     */
    hdmii_HPS.is_cable_detected = 0;
    hdmii_HPS.is_video_valid = 0;
    hdmii_HPS.is_unplugged = 0;

    fb_res_in.x   = 0;
    fb_res_in.y   = 0;
    fb_res_in.hfp = 0;
    fb_res_in.hsw = 0;
    fb_res_in.hbp = 0;
    fb_res_in.hsp = 0;
    fb_res_in.vfp = 0;
    fb_res_in.vsw = 0;
    fb_res_in.vbp = 0;
    fb_res_in.vsp = 0;



    // set getchar nonblocking
    tcgetattr(0, &initial_settings);

    new_settings              = initial_settings;
    new_settings.c_lflag      &= ~ICANON;
    new_settings.c_lflag      &= ~ECHO;
//     new_settings.c_lflag      &= ~ISIG;
    new_settings.c_cc[VMIN]   = 0;
    new_settings.c_cc[VTIME]  = 0;

    tcsetattr(0, TCSANOW, &new_settings);

    // register signal handler
    if ( signal(SIGINT, sig_handler) == SIG_ERR )  printf("Cannot catch SIGINT\n");

    //***********************************************************************************

    while (1) {

        /*
         * Check input cable plugged
         */
        check_hdmii_cable_plugged(&hdmii_HPS, &(fb.fmc_imageon), utia_hdmii_edid_content);

        /*
         * Check input cable unplugged
         */
        check_hdmii_cable_unplugged(&hdmii_HPS, &(fb.fmc_imageon));

        /*
         * Try to detect runtime resolution change on input inputs
         */
        check_hdmii_resolution(&hdmii_HPS, &fb, &fb_res_in);

        /*
         * configure VDMA
         */
        set_valid_hdmii_video_signal(&hdmii_HPS, &fb, &fb_res_in, memfd);

        /*
         * set offset
         */
        set_hdmio_offset(&hdmii_HPS, &fb, &fb_res_out, &fb_res_in, &offset_en, &offset );

        /*
         * print menu
         */
        ui_print_menu("Menu", menu, &menu_print_en);


        /*
         * GUI - terminal
         */
        keyin = getchar();

        if ( system_stop ) keyin = 'x';


        if (keyin != EOF) {

            key = keyin;

            switch(key) {

                case 'z':
                    offset_en = OFFSET_EN_ZERO;
                    break;
                case 'l':
                    offset_en = OFFSET_EN_LEFT;
                    break;
                case 'r':
                    offset_en = OFFSET_EN_RIGHT;
                    break;
                case 'd':
                    offset_en = OFFSET_EN_DOWN;
                    break;
                case 'u':
                    offset_en = OFFSET_EN_UP;
                    break;

                case '1':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1920x1080p60, &fb_res_out);
                    break;

                case '2':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1920x1200p50, &fb_res_out);
                    break;

                case '3':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1600x1200p50, &fb_res_out);
                    break;

                case '4':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1680x1050p60, &fb_res_out);
                    break;

                case '5':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1280x1024p60, &fb_res_out);
                    break;

                case '6':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1280x720p60, &fb_res_out);
                    break;

                case '7':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_1024x768p60 , &fb_res_out);
                    break;

                case '8':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_800x600p60 , &fb_res_out);
                    break;

                case '9':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_640x480p60 , &fb_res_out );
                    break;

                case 'a':
                    (void)set_resolution(&clkwiz, &vtc, &fb, VTC_600x800p60  , &fb_res_out);
                    break;


                case 'm':
                    menu_print_en = 1;
                    break;

                case 'x':
                    printf("Exiting\n");
                    tcsetattr(0, TCSANOW, &initial_settings);

                    if ( munmap(clkwiz_map.base, MAP_SIZE) == -1 ) {
                        printf("Clock Wizard userspace unmapping failed\n");
                        return XST_FAILURE;
                    }
                    if ( munmap(iic_map.base, MAP_SIZE) == -1 ) {
                        printf("Clock Wizard userspace unmapping failed\n");
                        return XST_FAILURE;
                    }
                    if ( munmap(vtc_map.base, MAP_SIZE) == -1 ) {
                        printf("Clock Wizard userspace unmapping failed\n");
                        return XST_FAILURE;
                    }
                    if ( munmap(vdma_map.base, MAP_SIZE) == -1 ) {
                        printf("Clock Wizard userspace unmapping failed\n");
                        return XST_FAILURE;
                    }

                    if ( close(memfd) ) {
                        printf("Cannot close /dev/mem\n");
                        return XST_FAILURE;
                    }

                    return XST_SUCCESS;
                    break;
            }
        }

    }

    return XST_FAILURE;
}
























